/*
 * AI Noise Reduction Example 
 * This firmware programs the SPU with a smarat noise cancellation algorithm that filters out background noise while preserving
 * human speech. Pressing the button turns on the algorithm, where the green LED indicates that the algorithm is on. Press the
 * button again to turn off the algorithm and hear the unfiltered audio output. The volume slider on the side of the Tympan controls
 * the output gain of the headphones. Various debugging information is also displayed in the serial terminal.
 *
 * Before running this firmware, do the following:
 * 1) Insert an SD card with the files 0PROG_A and 0PROG_D corresponding to the example AINR model.
 * 2) Wire the SPU to the Tympan according to the setup tutorial.
 * 3) Attach headphones to the black audio output jack on the Tympan.
 * 4) The Button2 library by Lennart Hennigs is required to run this firmware. Install it in the Arduino Library Manager.
 * 5) Ensure that AUDIO_BLOCK_SAMPLES in AudioStream.h is set to at least the value of HOP_SIZE below. This file is likely located at:
 * C:\Users\<USERNAME>\AppData\Local\Arduino15\packages\teensy\hardware\avr\1.57.2\cores\teensy4\AudioStream.h or similar
 *
 */

#include <Tympan_Library.h>    //Tympan library
#include <SD.h>                //SD card library
#include "ProcessAudio_F32.h"  //Audio processing algorithm
#include <AudioStream.h>       //Audio streaming library
#include "spu.h"               //Femtosense SPU library
#include <Button2.h>           //Button library

//Model parameters
#define HOP_SIZE 128  //I2S buffer size. You must update AUDIO_BLOCK_SAMPLES in AudioStream.h to a value of at least this size.
//For 16ms AINR model, use 128. For 8ms AINR model, use 64
//!!! This must also match HOP_SIZE in SPU.h !!!

//Audio Parameters
#define SAMPLE_RATE_HZ 16000.0f  //Other frequencies available listed in the table in AudioOutputI2S_F32
#define INPUT_GAIN_DB 25.0f      //gain on the microphone (0-47.5dB in 0.5dB steps)

//Audio input options:
//1) TYMPAN_INPUT_ON_BOARD_MIC //use the on board microphone
//2) TYMPAN_INPUT_JACK_AS_MIC //use the microphone jack - defaults to mic bias 2.5V
//3) TYMPAN_INPUT_JACK_AS_LINEIN //use the microphone jack - defaults to mic bias OFF
#define INPUT_SELECTION TYMPAN_INPUT_ON_BOARD_MIC

//IO pins for onboard LEDs
#define LED_GREEN 15
#define LED_RED 16
#define BUTTON 25

//IO pin for the flash chip SS on the SPU EVK board.
#define SS_FLASH 24

//IO pin for SPU Clock
#define SPU_CLK 6

//Serial BAUD rate
#define SERIAL_BAUD 115200


//configure audio
AudioSettings_F32 audio_settings(SAMPLE_RATE_HZ, HOP_SIZE);

//create audio library objects for handling the audio
Tympan myTympan(TympanRev::E, audio_settings);  //Initialize Tympan
AudioInputI2S_F32 i2s_in(audio_settings);       //Digital audio from Codec.
ProcessAudio_F32 effect1(audio_settings);       //Attach audio processing algorithm in ProcessAudio.h. This runs every AUDIO_BLOCK_SAMPLES2 samples.
AudioEffectGain_F32 gain1(audio_settings);      //This is for the volume knob
AudioOutputI2S_F32 i2s_out(audio_settings);     //Digital audio to Codec.

//audio connections
AudioConnection_F32 patchCord1(i2s_in, 0, effect1, 0);  //connect the Left input to the processing algorithm
AudioConnection_F32 patchCord2(effect1, 0, gain1, 0);   //connect to the gain effect
AudioConnection_F32 patchCord3(gain1, 0, i2s_out, 0);   //connect to the Left output
AudioConnection_F32 patchCord4(gain1, 0, i2s_out, 1);   //also connect to the Right output (mono -> stereo)

//setup button
Button2 button;

//current volume
float volume_dB = 0.0;  //this value will be overridden by volume knob position

//current loopback state
int loopback_mode = 1;  // 1=loopback mode, 0=AINR mode. Start in loopback mode.

//runs once at startup
void setup() {

  //Initialize serial port
  Serial.begin(SERIAL_BAUD);
  delay(1000);

  //Check that AUDIO_BLOCK_SAMPLES is set correctly
  if (AUDIO_BLOCK_SAMPLES < HOP_SIZE)
    error("You must increase AUDIO_BLOCK_SAMPLES in AudioStream.h to be >= HOP_SIZE", 2);

  //Disable flash chip that shares SPI bus with SPU
  pinMode(SS_FLASH, OUTPUT);
  digitalWrite(SS_FLASH, HIGH);

  //Turn on SPU clock
  analogWriteFrequency(SPU_CLK, 32000);
  analogWrite(SPU_CLK, 128);

  //Initialize SPU and run integrity and clock checks
  int init_error = initializeSPU();
  if (init_error == 0)
    error("SPU integrity check failed. Check wiring.", 3);
  if (init_error == -1)
    error("SPU clock check failed. Check clock jumper and signal.", 5);

  //Program SPU from SD Card
  if (!programSPU())
    error("Programming files cannot be read from SD card.", 4);

  //Start audio feed
  AudioMemory_F32(20, audio_settings);    //allocate the audio memory
  myTympan.enable();                      //activate Codec
  myTympan.inputSelect(INPUT_SELECTION);  //attach input
  myTympan.volume_dB(0);
  myTympan.setInputGain_dB(INPUT_GAIN_DB);  //set input gain

  //set initial volume from the volume slider
  servicePotentiometer(millis(), 0);

  //setup button
  button.begin(BUTTON);
  button.setTapHandler(tap);

  //Flash LED green to indicate device is ready
  flashLED(LED_GREEN, 3);
  Serial.println("Setup complete.");
}

//runs after setup and repeats forever.
void loop() {

  //handle volume change
  servicePotentiometer(millis(), 100);  //service the volume knob every 100ms

  //handle button press
  button.loop();
}

//Reads files 0PROG_A and 0PROG_D from root directory of SD card and programs SPU
//Returns 1 if successful, else returns 0
int programSPU() {

  //Mount SD card
  if (!SD.begin(BUILTIN_SDCARD)) {
    Serial.println("SD card initialization failed!");
    return 0;
  }

  //Setup files
  File addressFile;
  File dataFile;
  addressFile = SD.open("0PROG_A");
  dataFile = SD.open("0PROG_D");

  //Read file in chunks of 512 lines (this is faster than reading line by line)
  int i = 0;
  if (addressFile) {
    unsigned long addresses[512];
    while (1) {
      char address[9];
      addressFile.readBytes(address, 8);
      address[8] = '\0';
      char CR_LF = addressFile.read();  //consume CR or LF
      if (CR_LF == '\015')              //if CR, there will also be LF
        addressFile.read();             //consume LF

      addresses[i % 512] = strtoul(address, '\0', 16);

      if ((i % 512 == 511) || (!addressFile.available())) {

        //print every 512 lines to show progress
        Serial.print("Programming from file: Line ");
        Serial.print(i);
        Serial.println("...");

        //for each line in file
        for (int j = 0; j <= (i % 512); j++) {
          char data[9];
          dataFile.readBytes(data, 8);
          data[8] = '\0';
          CR_LF = dataFile.read();
          if (CR_LF == '\015')
            dataFile.read();
          unsigned long datas;
          datas = strtoul(data, '\0', 16);

          //write to SPU
          SPU_Write(0x01, (uint32_t)addresses[j], 1, (uint32_t*)&datas);
        }

        //file has ended
        if (!addressFile.available())
          break;
      }
      i++;
    }
    //could not find address file
  } else {
    Serial.println("Cannot find file 0PROG_A on SD card");
    return 0;
  }
  //completed successfully
  return 1;
}

//Flases LED flashCount number of times.
void flashLED(int LEDPin, int flashCount) {
  for (int i = 0; i < flashCount; i++) {
    digitalWrite(LEDPin, HIGH);
    delay(150);
    digitalWrite(LEDPin, LOW);
    delay(150);
  }
}

//Prints error message and ends code execution while flashing red LED flashCount number of times forever
void error(String errorMessage, int flashCount) {
  Serial.print("ERROR: ");
  Serial.println(errorMessage);

  //loop forever
  while (1) {
    flashLED(LED_RED, flashCount);
    delay(1000);
  }
}

//Checks position of volume knob and updates output gain accordingly. Updates are limited to once per updatePeriod_millis ms
void servicePotentiometer(unsigned long curTime_millis, unsigned long updatePeriod_millis) {

  //keep track of previous values
  static unsigned long lastUpdate_millis = 0;
  static float prev_val = 0;

  //has enough time passed to update everything?
  if (curTime_millis < lastUpdate_millis) lastUpdate_millis = 0;     //handle wrap-around of the clock
  if ((curTime_millis - lastUpdate_millis) > updatePeriod_millis) {  //is it time to update?

    //read potentiometer knob
    float val = float(myTympan.readPotentiometer()) / 1023.0;  //0.0 to 1.0
    val = (1.0 / 9.0) * (float)((int)(9.0 * val + 0.5));       //quantize so that it doesn't chatter...0 to 1.0

    //update volume is value has changed
    if (abs(val - prev_val) > 0.05) {  //is it different than before?
      prev_val = val;                  //save the value for comparison for the next time around

      //choose the desired gain value based on the knob setting
      const float min_gain_dB = -20.0, max_gain_dB = 40.0;          //set desired gain range
      volume_dB = min_gain_dB + (max_gain_dB - min_gain_dB) * val;  //computed desired gain value in dB

      //update volume via gain setting
      gain1.setGain_dB(volume_dB);  //set the gain
      Serial.print("Updating volume: Digital Gain dB = ");
      Serial.println(volume_dB);  //print text to Serial port
    }
    lastUpdate_millis = curTime_millis;
  }
}

//This is called whenever the button is pressed
void tap(Button2& button) {

  //toggle between loopback mode and AINR mode
  loopback_mode = (loopback_mode == 1) ? 0 : 1;
  effect1.setLoopbackMode(loopback_mode);

  //set green LED when AIRN mode is enabled
  digitalWrite(LED_GREEN, !loopback_mode);

  //print current mode to serial terminal
  if (loopback_mode) {
    Serial.println("Switched to Loopback Mode...");
  } else {
    Serial.println("Switched to AINR Mode...");
  }
}
