/* Wakeword Detection Example 
 * This firmware programs the SPU with a keyword spotting algorithm that listens for the phrase "Hey Snips." When the phrase
 * is heard, the red and green LEDs flash, and "Hey Snips!" is displayed on the serial terminal. Various debugging information
 * is also displayed in the serial terminal.
 *
 * Before running this firmware, do the following:
 * 1) Insert an SD card with the files 0PROG_A and 0PROG_D corresponding to the example WWD model.
 * 2) Wire the SPU to the Tympan according to the setup tutorial. Note: The button is not needed for this.
 * 3) Ensure that AUDIO_BLOCK_SAMPLES in AudioStream.h is set to at least the value of HOP_SIZE below. This file is likely located at:
 * C:\Users\<USERNAME>\AppData\Local\Arduino15\packages\teensy\hardware\avr\1.57.2\cores\teensy4\AudioStream.h or similar
 *
 */

#include <Tympan_Library.h>    //Tympan library
#include <SD.h>                //SD card library
#include "ProcessAudio_F32.h"  //Audio processing algorithm
#include <AudioStream.h>       //Audio streaming library
#include "spu.h"               //Femtosense SPU library

//Model parameters
#define HOP_SIZE 256  //I2S buffer size. You must update AUDIO_BLOCK_SAMPLES in AudioStream.h to a value of at least this size.
//!!! This must also match HOP_SIZE in SPU.h !!!

//Audio Parameters
#define SAMPLE_RATE_HZ 16000.0f  //Other frequencies available listed in the table in AudioOutputI2S_F32
#define INPUT_GAIN_DB 47.5f      //gain on the microphone (0-47.5dB in 0.5dB steps)

//Audio input options:
//1) TYMPAN_INPUT_ON_BOARD_MIC //use the on board microphone
//2) TYMPAN_INPUT_JACK_AS_MIC //use the microphone jack - defaults to mic bias 2.5V
//3) TYMPAN_INPUT_JACK_AS_LINEIN //use the microphone jack - defaults to mic bias OFF
#define INPUT_SELECTION TYMPAN_INPUT_ON_BOARD_MIC

//IO pins for onboard LEDs
#define LED_GREEN 15
#define LED_RED 16

//IO pin for the flash chip SS on the SPU EVK board.
#define SS_FLASH 24

//IO pin for SPU Clock
#define SPU_CLK 6

//Serial BAUD rate
#define SERIAL_BAUD 115200


//configure audio
AudioSettings_F32 audio_settings(SAMPLE_RATE_HZ, HOP_SIZE);

//create audio library objects for handling the audio
Tympan myTympan(TympanRev::E, audio_settings);  //Initialize Tympan
AudioInputI2S_F32 i2s_in(audio_settings);       //Digital audio from Codec.
ProcessAudio_F32 effect1(audio_settings);       //Attach audio processing algorithm in ProcessAudio.h. This runs every AUDIO_BLOCK_SAMPLES2 samples.

//connect the Left input to the processing algorithm
AudioConnection_F32 patchCord1(i2s_in, 0, effect1, 0);

//runs once at startup
void setup() {

  //Initialize serial port
  Serial.begin(SERIAL_BAUD);
  delay(1000);

  //Check that AUDIO_BLOCK_SAMPLES is set correctly
  if (AUDIO_BLOCK_SAMPLES < HOP_SIZE)
    error("You must increase AUDIO_BLOCK_SAMPLES in AudioStream.h to be >= HOP_SIZE", 2);

  //Disable flash chip that shares SPI bus with SPU
  pinMode(SS_FLASH, OUTPUT);
  digitalWrite(SS_FLASH, HIGH);

  //Turn on SPU clock
  analogWriteFrequency(SPU_CLK, 32000);
  analogWrite(SPU_CLK, 128);

  //Initialize SPU and run integrity and clock checks
  int init_error = initializeSPU();
  if (init_error == 0)
    error("SPU integrity check failed. Check wiring.", 3);
  if (init_error == -1)
    error("SPU clock check failed. Check clock jumper and signal.", 5);

  //Program SPU from SD Card
  if (!programSPU())
    error("Programming files cannot be read from SD card.", 4);

  //Start audio feed
  AudioMemory_F32(20, audio_settings);      //allocate the audio memory
  myTympan.enable();                        //activate Codec
  myTympan.inputSelect(INPUT_SELECTION);    //attach input
  myTympan.setInputGain_dB(INPUT_GAIN_DB);  //set input gain

  //Flash LED green to indicate device is ready
  flashLED(LED_GREEN, 3);
  Serial.println("Setup complete.");
}

//Reads files 0PROG_A and 0PROG_D from root directory of SD card and programs SPU
//Returns 1 if successful, else returns 0
int programSPU() {

  //Mount SD card
  if (!SD.begin(BUILTIN_SDCARD)) {
    Serial.println("SD card initialization failed!");
    return 0;
  }

  //Setup files
  File addressFile;
  File dataFile;
  addressFile = SD.open("0PROG_A");
  dataFile = SD.open("0PROG_D");

  //Read file in chunks of 512 lines (this is faster than reading line by line)
  int i = 0;
  if (addressFile) {
    unsigned long addresses[512];
    while (1) {
      char address[9];
      addressFile.readBytes(address, 8);
      address[8] = '\0';
      char CR_LF = addressFile.read();  //consume CR or LF
      if (CR_LF == '\015')              //if CR, there will also be LF
        addressFile.read();             //consume LF

      addresses[i % 512] = strtoul(address, '\0', 16);

      if ((i % 512 == 511) || (!addressFile.available())) {

        //print every 512 lines to show progress
        Serial.print("Programming from file: Line ");
        Serial.print(i);
        Serial.println("...");

        //for each line in file
        for (int j = 0; j <= (i % 512); j++) {
          char data[9];
          dataFile.readBytes(data, 8);
          data[8] = '\0';
          CR_LF = dataFile.read();
          if (CR_LF == '\015')
            dataFile.read();
          unsigned long datas;
          datas = strtoul(data, '\0', 16);

          //write to SPU
          SPU_Write(0x01, (uint32_t)addresses[j], 1, (uint32_t*)&datas);
        }

        //file has ended
        if (!addressFile.available())
          break;
      }
      i++;
    }
    //could not find address file
  } else {
    Serial.println("Cannot find file 0PROG_A on SD card");
    return 0;
  }
  //completed successfully
  return 1;
}

//Flases LED flashCount number of times.
void flashLED(int LEDPin, int flashCount) {
  for (int i = 0; i < flashCount; i++) {
    digitalWrite(LEDPin, HIGH);
    delay(150);
    digitalWrite(LEDPin, LOW);
    delay(150);
  }
}

//Prints error message and ends code execution while flashing red LED flashCount number of times forever
void error(String errorMessage, int flashCount) {
  Serial.print("ERROR: ");
  Serial.println(errorMessage);

  //loop forever
  while (1) {
    flashLED(LED_RED, flashCount);
    delay(1000);
  }
}

//runs after setup and repeats forever.
void loop() {
  ;
}
