#include "spu.h"

/*Initializes the SPU
 *returns 0 if registers cannot be read
 *returns -1 if PLL does not lock
 *returns 1 if everything is okay
 */
int initializeSPU() {

  //initialize SPI on default pins
  SPI.begin();

  //setup reset and SPI SS pins
  pinMode(SPU_RESET_PIN, OUTPUT);
  pinMode(SS, OUTPUT);

  //reset SPU
  SPUReset(SPU_RESET_PIN);

  //Check that SPU is responsive, return 0 if not.
  if (!SPUIntegrityCheck())
    return 0;

  //setup SPU PLL, return -1 if PLL lock failure
  if (!SPUPLLSetup())
    return -1;

  //Manually turn on memory banks
  //Note: neccessary memory banks area already turned on automatically in default programming files
  //SPUMemorySetup();

  return 1;
}

//Resets the SPU using the reset pin
void SPUReset(int resetPin) {
  Serial.print("Resetting SPU...");

  digitalWrite(resetPin, HIGH);
  delay(1000);
  digitalWrite(resetPin, LOW);
  delay(1000);

  Serial.println("reset.");
}

//Sets up the SPU PLL registers
int SPUPLLSetup() {

  //buffer to write to SPU
  uint32_t SPUBuffer[4];

  //Change clock divider to 100MHz for 32KHz reference clock
  uint32_t CLKF = 0xC34;
  uint32_t CLKOD = 0x0;
  uint32_t CLKR = 0x0;
  uint32_t BWADJ = 0x61A;

  //these fields are offset by 1 bit
  SPUBuffer[0] = BWADJ << 1;
  SPUBuffer[1] = CLKOD << 1;
  SPUBuffer[2] = CLKF << 1;
  SPUBuffer[3] = CLKR << 1;
  SPU_Write(0x00, 12, 4, SPUBuffer);

  //turn off core 1 if not needed
  /*SPUBuffer[0] = 1 << 2;
  SPU_Write(0x00, 0, 1, SPUBuffer);*/

  //enable oscillator instead of reference clock
  //default: 248
  //osc on: 448
  /*SPUBuffer[0] = 0x448;
  SPU_Write(0x00, 4, 1, SPUBuffer);*/

  //turn off PLL
  /*SPUBuffer[0] = (1 << 5);
  SPU_Write(0x00, 8, 1, SPUBuffer);*/

  //reset PLL to update
  SPUBuffer[0] = 1;
  SPU_Write(0x00, 8, 1, SPUBuffer);
  SPUBuffer[0] = 0;
  SPU_Write(0x00, 8, 1, SPUBuffer);

  //wait for PLL to lock
  delay(1000);

  //check PLL lock 
  SPU_Read(0x00, 28, 1, SPUBuffer);
  return SPUBuffer[0] & 0x00000001;

}

//Turns on all of the SPU memory banks
void SPUMemorySetup() {

  //buffer to write to SPU
  uint32_t SPUBuffer[26];

  //first memory core conf register location
  uint32_t MemPwrBaseAddr = 0x34;

  // put memories in CD state
  for (int i = 0; i < 26; i++) {
    SPUBuffer[i] = 0x05098;
  }
  SPU_Write(0x1, MemPwrBaseAddr, 26, SPUBuffer);

  //turn on each memory
  for (int i = 0; i < 26; i++) {
    SPUBuffer[i] = 0x11098;
  };
  SPU_Write(0x1, MemPwrBaseAddr, 26, SPUBuffer);
}

/* Request to read SPU data.
 * instruction: SPU instruction 
 * address: start address of data requested
 * length: length of data requested
 * data: buffer for returned data from SPU
 */
void SPU_Read(byte instruction, uint32_t address, int length, uint32_t* data) {
  int MessageLength = 1 + 8 + 4;    //bytes required for request
  byte WriteBuffer[MessageLength];  //holds instruction, start/end addresses, and 4 stall bytes

  WriteBuffer[0] = (instruction << 1);  //set wr_en to 0 for reading

  //set start and end addresses
  for (int i = 0; i < 4; i++) {
    WriteBuffer[1 + i] = (byte)((address << (24 - i * 8)) >> 24);
    WriteBuffer[5 + i] = (byte)(((address + length * 4) << (24 - i * 8)) >> 24);
  }
  WriteBuffer[MessageLength - 4] = 0x00;  //stall bytes
  WriteBuffer[MessageLength - 3] = 0x00;  //stall bytes
  WriteBuffer[MessageLength - 2] = 0x00;  //stall bytes
  WriteBuffer[MessageLength - 1] = 0x00;  //stall bytes

  //SPI transaction to request read
  SPI.beginTransaction(SPISettings(SPI_SPEED, MSBFIRST, SPI_MODE0));
  digitalWrite(SS, LOW);                     //assert SPU SS
  SPI.transfer(WriteBuffer, MessageLength);  //no response expected for write, so no SPI receive data buffer is set
  digitalWrite(SS, HIGH);                    //deassert SPU SS
  SPI.endTransaction();

  //setup read buffer to receive data
  MessageLength = (4 * length);
  byte ReadBuffer[MessageLength];

  //SPI transaction to read data
  SPI.beginTransaction(SPISettings(SPI_SPEED, MSBFIRST, SPI_MODE0));
  digitalWrite(SS, LOW);                    //assert SPU SS
  SPI.transfer(ReadBuffer, MessageLength);  //response will be written to ReadBuffer. Anything can be written here, so no SPI send data buffer is set
  digitalWrite(SS, HIGH);                   //deassert SPU SS
  SPI.endTransaction();

  //store data that was read into function output buffer
  for (int i = 0; i < length; i++) {
    data[i] = ReadBuffer[i * 4 + 0] + ((uint32_t)ReadBuffer[i * 4 + 1] << 8) + ((uint32_t)ReadBuffer[i * 4 + 2] << 16) + ((uint32_t)ReadBuffer[i * 4 + 3] << 24);
  }
}

/* Write data to SPU.
 * instruction: SPU instruction 
 * address: start address of write location
 * length: length of data to write
 * data: buffer of data to write to SPU
 */
void SPU_Write(byte instruction, uint32_t address, int length, uint32_t* data) {

  int MessageLength = (1 + 8 + 4 * length + 2);  //bytes required for request
  byte WriteBuffer[MessageLength];               //holds instruction, start/end addresses, data, and 2 stall bytes

  WriteBuffer[0] = (instruction << 1) | 0x01;  //set wr_en to 1 for writing

  //set start and end addresses
  for (int i = 0; i < 4; i++) {
    WriteBuffer[1 + i] = (byte)((address << (24 - i * 8)) >> 24);
    WriteBuffer[5 + i] = (byte)(((address + length * 4) << (24 - i * 8)) >> 24);
  }

  //load data into write buffer
  for (int i = 0; i < length; i++) {
    WriteBuffer[9 + i * 4 + 0] = (byte)((data[i] << 24) >> 24);
    WriteBuffer[9 + i * 4 + 1] = (byte)((data[i] << 16) >> 24);
    WriteBuffer[9 + i * 4 + 2] = (byte)((data[i] << 8) >> 24);
    WriteBuffer[9 + i * 4 + 3] = (byte)(data[i] >> 24);
  }
  WriteBuffer[MessageLength - 2] = 0x00;  //stall bytes
  WriteBuffer[MessageLength - 1] = 0x00;  //stall bytes

  //SPI transaction to write data
  SPI.beginTransaction(SPISettings(SPI_SPEED, MSBFIRST, SPI_MODE0));
  digitalWrite(SS, LOW);                     //assert SPU SS
  SPI.transfer(WriteBuffer, MessageLength);  //no response expected for write, so no SPI receive data buffer is set
  digitalWrite(SS, HIGH);                    //deassert SPU SS
  SPI.endTransaction();
}

//Tests basic SPU register reads
int SPUIntegrityCheck() {

  //simply check that we can read the default values of the config registers
  uint32_t SPUBuffer[7];
  uint32_t SPUDefaults[7] = { 0x00, 0x248, 0x00, 0x3E6, 0x02, 0x7CE, 0x00 };

  //read values from SPU
  SPU_Read(0x00, 0, 7, SPUBuffer);

  //compare to defaults
  int OKAY = 1;
  for (int i = 0; i < 7; i++)
    if (SPUBuffer[i] != SPUDefaults[i])
      OKAY = 0;

  return OKAY;
}
